/*
 * $Id: IParameterClient.java,v 1.5 2006/11/22 10:20:46 wzb Exp $
 *
 * $Log: IParameterClient.java,v $
 * Revision 1.5  2006/11/22 10:20:46  wzb
 * Final final version for 5.5.3
 *
 * Revision 1.2.2.1  2006/11/06 15:02:47  wzb
 * R5.5.3 bugfixes and additions
 *
 *
 */
package esa.span.client;

import esa.span.common.PrmReport;

/**
 * This interface can be used to receive events for parameter updates
 * coming from the Spectrum Analyser. Implementing classes must deliver
 * a list of parameter names at startup.
 * 
 * @author Wernke zur Borg
 *
 */
public interface IParameterClient extends IMonitoringClient
{
	/**
	 * This retrieves the names of all parameters the client wants
	 * to get updates for. It is only called once at system startup, 
	 * so the set of monitored parameters cannot be changed later on.
	 *  
	 * @return  A list of parameter names the client wants to receive 
	 * updates for. Every parameter must be in the form as required 
	 * by the Spectrum Analyser for querying the value, i.e. including
	 * the trailing '?'.
	 */
	public String[] getPrmNames();
	
	/**
	 * This delivers a new value for a parameter.
	 * 
	 * @param prmName  The parameter name.
	 * @param value  The current parameter value.
	 */
	public void updatePrmValue( String prmName, PrmReport value);
}
