/*
 * $Id: SpanDisplay.java,v 1.7 2007/03/06 13:35:23 wzb Exp $
 *
 * $Log: SpanDisplay.java,v $
 * Revision 1.7  2007/03/06 13:35:23  wzb
 * R5.5.4 bugfixes for final delivery
 *
 * Revision 1.6  2006/11/22 10:20:46  wzb
 * Final final version for 5.5.3
 *
 * Revision 1.3.2.3  2006/11/20 13:08:42  wzb
 * STC R5.5.3 final
 *
 * Revision 1.3.2.2  2006/11/06 15:02:48  wzb
 * R5.5.3 bugfixes and additions
 *
 *
 */
package esa.span.client;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Hashtable;
import java.util.Stack;
import java.util.TimeZone;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYPolygonAnnotation;
import org.jfree.chart.annotations.XYTextAnnotation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.event.ChartChangeEvent;
import org.jfree.chart.event.ChartProgressEvent;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.Range;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.RectangleInsets;
import org.jfree.ui.TextAnchor;

import esa.span.common.PrmReport;
import esa.span.common.SpanType;
import esa.span.common.VarValueType;

/**
 * This class displays the spectrum plot with the means of its
 * parent class, {@link ChartPanel} and other classes in the JFreeChart
 * library.
 * @author Wernke zur Borg
 *
 */
public class SpanDisplay extends ChartPanel	implements IParameterClient, ITraceClient
{
	private static final long serialVersionUID = -4569747405267911705L;

//	private static int mySeq = 0;
	private Stack<Rectangle2D> zoomStack = new Stack<Rectangle2D>();

	private final String CMD_SNAPSHOT = "SNAPSHOT";
	private final String CMD_QUICK_PRINT = "QUICK_PRINT";
	private final String CMD_ZOOM_BACK = "ZOOM_BACK";
	private final String CMD_ZOOM_OUT = "ZOOM_OUT";
	private final String CMD_ANTI_ALIAS = "ANTI_ALIAS";
	
	// A hash map with all parameters subscribed to
	private Hashtable<String, PrmReport> myPrmData = new Hashtable<String, PrmReport>();
	
	private final String PRM_RLEV   = "DISP:WIND:TRAC:Y:RLEV?";
	private final String PRM_PDIV   = "DISP:WIND:TRAC:Y:PDIV?";
	private final String PRM_STAR   = "FREQ:STAR?";
	private final String PRM_SPAN   = "FREQ:SPAN?";
	private final String PRM_MARKER = "CALC:MARK:MODE?";
	private final String PRM_MARKX  = "CALC:MARK:X?";
	private final String PRM_MARKY  = "CALC:MARK:Y?";
	private final String PRM_SWPTIME = "SWE:TIME?";
	
	private boolean crosshairVisible = false;
//	private boolean yCrosshairLockedOnData = true;
	
	private Range xAxisRange = null;
	private Range yAxisRange = null;
	private double xAxisFactor = 1.0;
	
	private boolean zeroSpanMode = false;
	
	/** date formatter for timestamp */
	XYTextAnnotation timeStampAnnotation;
	final SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
	
    /** A horizontal trace line. */
    private transient Line2D horizontalTraceLine;

	/*
	 *  Do not initialise the following two variables - their value
	 *  is set in the overloaded function createPopupMenu(), which is called
	 *  by the constructor of the baseclass, and this happens *before* the
	 *  initialisation of local object variables. So do not even set them
	 *  to 'null' !
	 */
    private JCheckBoxMenuItem toggleChartCheckBox;
    private JMenuItem snapshotMenuItem;
    private JMenuItem quickPrintMenuItem;
	private JMenuItem zoomBackMenuItem;
	private JMenuItem zoomOutMenuItem;
	private JCheckBoxMenuItem antiAliasCheckBox;
	private AbstractAction toggleCursorsAction;
	

	/* (non-Javadoc)
	 * @see org.jfree.chart.ChartPanel#restoreAutoBounds()
	 */
	public void restoreAutoBounds() 
	{
		// disabled because of strange effects when dragging the mouse from
		// lower right to upper left. See ChartPanel.mouseReleased().
	}

	/**
	 * Constructor.
	 */
	public SpanDisplay()
    {
        super( createSpanChart(), 
                DEFAULT_WIDTH,
                DEFAULT_HEIGHT,
                DEFAULT_MINIMUM_DRAW_WIDTH,
                DEFAULT_MINIMUM_DRAW_HEIGHT,
                DEFAULT_MAXIMUM_DRAW_WIDTH,
                DEFAULT_MAXIMUM_DRAW_HEIGHT,
        		true,	// double buffer
        		false,	// properties
        		false,	// save
        		true,	// print
        		false,	// zoom
        		true	// tooltips
        );

		try
        {
            initialize();
        } 
		catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }

    @SuppressWarnings("deprecation")
	private void initialize() throws Exception
    {
        JFreeChart chart = getChart();
        chart.setAntiAlias(false);
        chart.setPadding(new RectangleInsets(5,0,0,8));
        
        // just an arbitrary starting range
        xAxisRange = new Range(69.5, 70.5);
        yAxisRange = new Range(-120, -20);
        
        sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
        
        Font axisFont = new Font("SansSerif", Font.PLAIN, 13);
        RectangleInsets axisLabelInsets = 
        	new RectangleInsets(2.0, 2.0, 0.0, 2.0);
        
        XYPlot plot = chart.getXYPlot();
        ValueAxis xAxis = plot.getDomainAxis();
        plot.setBackgroundPaint(Color.black);
        xAxis.setRange(xAxisRange);
        xAxis.setLabelFont(axisFont);
        xAxis.setTickLabelFont(axisFont);
        xAxis.setLabelInsets(axisLabelInsets);
        xAxis.setTickLabelInsets(axisLabelInsets);
        xAxisFactor = 1000000.0;
        plot.getRenderer().setPaint(Color.yellow);
        ValueAxis yAxis = plot.getRangeAxis();
        yAxis.setRange(yAxisRange);
        yAxis.setLabelFont(axisFont);
        yAxis.setTickLabelFont(axisFont);
        yAxis.setLabelInsets(axisLabelInsets);
        yAxis.setTickLabelInsets(axisLabelInsets);
        plot.setDomainCrosshairLockedOnData(false);
		plot.setDomainCrosshairPaint(Color.red);
		plot.setDomainCrosshairStroke(new BasicStroke(1.5f));
		plot.setRangeCrosshairPaint(Color.red);
		plot.setRangeCrosshairStroke(new BasicStroke(1.5f));
        plot.setRangeCrosshairLockedOnData(false);

//        this.setHorizontalTraceLine(new Line2D());
        this.setVerticalAxisTrace(crosshairVisible);
        this.setHorizontalAxisTrace(crosshairVisible);
//        this.setFillZoomRectangle(true);
		plot.setDomainCrosshairVisible(crosshairVisible);
		plot.setRangeCrosshairVisible(crosshairVisible);
		
		resetAxesRange();
		
		timeStampAnnotation = new XYTextAnnotation("", 0, 0);
    	timeStampAnnotation.setFont(new Font("Dialog", Font.PLAIN, 12));
    	timeStampAnnotation.setTextAnchor(TextAnchor.BOTTOM_RIGHT);
    	timeStampAnnotation.setBackgroundPaint(Color.DARK_GRAY);
    	timeStampAnnotation.setPaint(Color.WHITE);
        
    	// start trace 1 automatically
    	SpanClient.getClient().addTraceClient((short)1, this);
    	SpanClient.getClient().addPrmClient(this);
    	SpanClient.getClient().addSpanModeListener(this);
    	
//    	getChart().addProgressListener(this);
    }
    
    /* (non-Javadoc)
     * @see org.jfree.chart.ChartPanel#actionPerformed(java.awt.event.ActionEvent)
     */
    public void actionPerformed(ActionEvent event)
	{
    	String cmd = event.getActionCommand();
    	if(cmd.equals(CMD_SNAPSHOT))
    	{
    		SpanMainFrame.getMainFrame().windowDump();
    	}
    	else if(cmd.equals(CMD_ZOOM_OUT))
    	{
    		zoomOut();
    	}
    	else if(cmd.equals(CMD_ZOOM_BACK))
    	{
    		zoomBack();
    	}
    	else if(cmd.equals(CMD_ANTI_ALIAS))
    	{
    		getChart().setAntiAlias(antiAliasCheckBox.isSelected());
    	}
    	else if(cmd.equals(CMD_QUICK_PRINT))
    	{
    		SpanMainFrame.getMainFrame().windowPrint(false);
    	}
    	else if(cmd.equals(PRINT_COMMAND)) 
    	{
    		SpanMainFrame.getMainFrame().windowPrint(true);
    	}
    	else
    	{
    		super.actionPerformed(event);
    	}
    	
	}

	/**
	 * Displays a dataset in the plot area.
	 * @param index the index of the dataset
	 * @param ymin the minimum y value
	 * @param ymax the maximum y value
	 * @param dataset the data set
	 * @param xAxisLabel the label to be attached to the x axis
	 */
	public void setDataset( int index, double ymin, double ymax, 
			XYDataset dataset, String xAxisLabel )
    {
		XYPlot plot = getChart().getXYPlot();

		if( dataset.getItemCount(0) < 1 )
		{
			plot.setDataset( index, null );
			return;
		}
		plot.getDomainAxis().setLabel(xAxisLabel);
		plot.setDataset( index, dataset );

		double x1 = dataset.getXValue(0, 0);
		double x2 = dataset.getXValue(0, dataset.getItemCount(0)-1);
		Range xRange = new Range(x1,x2);
		if( !xRange.equals(xAxisRange) )
		{
			xAxisRange = xRange;
			if(!isZoomed())
				resetXAxisRange();
		}
		
		Range yRange = new Range(ymin,ymax); 
		if( !yRange.equals(yAxisRange))
		{
			yAxisRange = yRange;
			if(!isZoomed())
				resetYAxisRange();
		}
    }

	private void resetXAxisRange()
	{
		NumberAxis axis = (NumberAxis)getChart().getXYPlot().getDomainAxis();
		axis.setRange(xAxisRange);
		// divide X range into 10 equal pieces
		double delta = (xAxisRange.getUpperBound() - xAxisRange.getLowerBound())/10;
		axis.setTickUnit(new NumberTickUnit(delta));
	}
	
	private void resetYAxisRange()
	{
		NumberAxis axis = (NumberAxis)getChart().getXYPlot().getRangeAxis();
		axis.setRange(yAxisRange);
		// divide Y range into 10 equal pieces
		double delta = (yAxisRange.getUpperBound() - yAxisRange.getLowerBound())/10;
		axis.setTickUnit(new NumberTickUnit(delta));
	}
	
	private void resetAxesRange()
	{
		resetXAxisRange();
		resetYAxisRange();
	}
	
    /* (non-Javadoc)
     * @see org.jfree.chart.ChartPanel#chartProgress(org.jfree.chart.event.ChartProgressEvent)
     */
    public void chartProgress(ChartProgressEvent event)
    {
//    	System.out.println("Chart progress: " + event.getType() 
//    			+ " / " + event.getPercent() + "%");
    	super.chartProgress(event);
//    	adjustCrosshairs(event.getChart().getXYPlot(), 1);
    	
    }
    
    /* (non-Javadoc)
     * @see org.jfree.chart.ChartPanel#chartChanged(org.jfree.chart.event.ChartChangeEvent)
     */
    public void chartChanged(ChartChangeEvent event)
    {
//    	System.out.println("Chart changed: " + event.getType().toString());
    	super.chartChanged(event);
//    	JFreeChart chart = event.getChart();
//    	if( chart != null )
//    	{
//    		XYPlot plot = chart.getXYPlot();
//    		if( plot != null )
//    		{
//    			adjustCrosshairs(event.getChart().getXYPlot(), 1);
//    		}
//    	}
    }
    
	/**
	 * Creates the SPAN chart.
	 * @return the SPAN chart
	 */
	static public JFreeChart createSpanChart()
    {
        return ChartFactory.createXYLineChart
                (null, // Title
                 "Frequency [MHz]", // X-Axis label
                 "Amplitude [dBm]", // Y-Axis label
                 null, // Dataset
                 PlotOrientation.VERTICAL,
                 false, // Show legend
                 true, // tooltips
                 false // URLs
                );
    }

    /* (non-Javadoc)
     * @see esa.span.client.ITraceClient#traceUpdate(short, double[])
     */
    public synchronized void traceUpdate( short traceId, double[] traceData, long timestamp )
    {
    	// workaround to avoid strange shapes of zooming rectangles
    	if( isDraggingForZoom() ) {
    		return;
    	}
    	
	    try
	    {
	    	PrmReport pr = getPrmReport(PRM_SPAN);
	    	if(pr.isValid()) {
	    		double spanx = pr.prmValue.realValue();

	    		if( spanx > 0.1 )
	    		{
	    			normalTraceUpdate( traceId, traceData, spanx, timestamp );
	    		}
	    		else
	    		{
	    			zeroSpanTraceUpdate( traceId, traceData, timestamp );
	    		}
	    	} else {
	    		// span value not available
	    	}
	    }
	    catch(Exception ex)
	    {
	    	SpanAppl.getLogger().error("Exception while processing traceUpdate", ex);
	    	ex.printStackTrace();
	    }
    }
    
    private void normalTraceUpdate( short traceId, double[] traceData, double spanx, long timestamp )
    	throws Exception
    {
    	setZeroSpanMode(false);
    	
    	double yrlev = 0.0;
    	double ypdiv = 0.0;
    	double startx = 0.0;
    	double markx =  0.0;
    	double marky =  0.0;
    	
	    TraceDataSeries data = new TraceDataSeries("");
	    
	    yrlev =  getValidValue(PRM_RLEV).realValue();
	    ypdiv =  getValidValue(PRM_PDIV).realValue();
	    startx = getValidValue(PRM_STAR).realValue();
	    markx =  getValidValue(PRM_MARKX).realValue();
	    marky =  getValidValue(PRM_MARKY).realValue();

    	double ymin = yrlev - ypdiv*10.0;
    	String markerMode = "OFF";
    	
    	try
    	{
    		markerMode = getValidValue(PRM_MARKER).visibleStringValue();
    	}
    	catch(Exception ex) {
//    		SpanAppl.getLogger().error("Exception while processing normalTraceUpdate", ex);
    	}

    	xAxisFactor = 1.0;
    	// adjust x axis labels
    	String[] units = {"Hz", "kHz", "MHz", "GHz"};
    	int index = 0;
    	double x = Math.abs(startx);
    	while( x > 1000.0 && index<units.length-1 )
    	{
    		xAxisFactor *= 1000.0;
    		x /= 1000.0;
    		index++;
    	}

    	startx /= xAxisFactor;
    	spanx  /= xAxisFactor;
    	markx  /= xAxisFactor;

    	double stepx = spanx/(traceData.length - 1);
    	for( int i=0; i<traceData.length; i++ )
    	{
    		data.add( startx+i*stepx, traceData[i], false );
    	}
    	
    	setDataset( traceId, ymin, yrlev, new XYSeriesCollection(data), "Frequency [" + units[index] + "]" );

    	XYPlot plot = getChart().getXYPlot();
    	plot.clearAnnotations();
		ValueAxis xAxis = plot.getDomainAxis();
		ValueAxis yAxis = plot.getRangeAxis();
 
    	if(markerMode.equals("POS") || 
    			SpanClient.getClient().getSpanType() == SpanType.HP)
    	{
    		double[] polygon = new double[8];
    		double xd = (xAxis.getUpperBound() - xAxis.getLowerBound())/getWidth()*5;
    		double yd = (yAxis.getUpperBound() - yAxis.getLowerBound())/getHeight()*5;
    		polygon[0] = markx + xd;
    		polygon[1] = marky;
    		polygon[2] = markx;
    		polygon[3] = marky - yd;
    		polygon[4] = markx - xd;
    		polygon[5] = marky;
    		polygon[6] = markx;
    		polygon[7] = marky + yd;
    		XYPolygonAnnotation pa = new XYPolygonAnnotation(
    				polygon, new BasicStroke(1.5f), Color.white);
    		plot.addAnnotation(pa);
    	}
    	
    	// add timestamp
    	timeStampAnnotation.setText(sdf.format(new Date(timestamp)));
    	timeStampAnnotation.setX(xAxis.getUpperBound());
    	timeStampAnnotation.setY(yAxis.getLowerBound());
    	plot.addAnnotation(timeStampAnnotation);
    }

    private void zeroSpanTraceUpdate( short traceId, double[] traceData, long timestamp )
    	throws Exception
    {
    	setZeroSpanMode(true);
    	
	    TraceDataSeries data = new TraceDataSeries("");

	    double yrlev =  getValidValue(PRM_RLEV).realValue();
    	double ypdiv =  getValidValue(PRM_PDIV).realValue();
    	double ymin = yrlev - ypdiv*10.0;

    	double spanx = getValidValue(PRM_SWPTIME).realValue();

    	xAxisFactor = 1.0;
    	// adjust x axis labels
    	String[] units = {"s", "ms", "\u00B5s"};
    	int index = 0;
    	while( spanx < 1.0 && index<units.length-1 )
    	{
    		xAxisFactor /= 1000.0;
    		spanx *= 1000.0;
    		index++;
    	}

    	double stepx = spanx/(traceData.length - 1);
    	for( int i=0; i<traceData.length; i++ )
    	{
    		data.add( i*stepx, traceData[i], false );
    	}

    	setDataset( traceId, ymin, yrlev, new XYSeriesCollection(data), 
    			"Sweep time [" + units[index] + "]" );

    	//System.out.println("Trace data (" + (++mySeq) + ") updated");

    	XYPlot plot = getChart().getXYPlot();
    	plot.clearAnnotations();
    }
    
    /* (non-Javadoc)
     * @see esa.span.client.IParameterClient#getPrmNames()
     */
    public String[] getPrmNames()
    {
    	String[] names = {PRM_PDIV, PRM_RLEV, PRM_SPAN, PRM_STAR, 
    			PRM_MARKER, PRM_MARKX, PRM_MARKY, PRM_SWPTIME};
    	return names;
    }
    
    /* (non-Javadoc)
     * @see esa.span.client.IParameterClient#updatePrmValue(java.lang.String, esa.span.common.PrmReport)
     */
    public synchronized void updatePrmValue(String prmName, PrmReport value)
    {
    	myPrmData.put(prmName, value);
    }

    private PrmReport getPrmReport( String name ) throws Exception
	{
		if( myPrmData.containsKey(name))
			return myPrmData.get(name);
		
		throw new Exception("No value for parameter " + name);
	}
    
    private VarValueType getValidValue(String name) throws Exception {
    	PrmReport rep = myPrmData.get(name);
    	if(rep != null) {
    		if(rep.isValid()) {
    			return rep.prmValue;
    		}
    	}
		
		throw new Exception("No valid value for parameter " + name);
    }

	/* (non-Javadoc)
	 * @see esa.span.client.SpanModeListener#spanModeChanged(esa.span.client.SpanModeListener.SpanMode)
	 */
	public void spanModeChanged(SpanMode newMode) 
	{
		if( newMode == SpanMode.MODE_DISCONNECTED )
		{
			clearPlots();
		}
	}
	
    /**
     * Clears all plots.
     */
    public void clearPlots() 
	{
    	for( int i=1; i<SpanClient.SPAN_NUM_TRACES; i++ )
    	{
    		getChart().getXYPlot().setDataset(i, null);
    	}
	}

	/**
	 * Sets the visibility of the crosshair cursor.
	 * @param b true=visible false=invisible
	 */
	public void setCursorVisible(boolean b)
	{
		if( b != crosshairVisible )
		{
			crosshairVisible = b;

			// reset crosshair position if cursor is switched on or off
			XYPlot plot = getChart().getXYPlot();
			plot.setDomainCrosshairValue(0.0);
			plot.setRangeCrosshairValue(0.0);
		}
  	}

	/* (non-Javadoc)
	 * @see org.jfree.chart.ChartPanel#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e) {
		super.mouseClicked(e);
		if(e.getClickCount() == 2) {
			toggleChartCheckBox.doClick();
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jfree.chart.ChartPanel#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
		super.mouseEntered(e);
    	XYPlot plot = getChart().getXYPlot();
    	plot.setDomainCrosshairVisible(crosshairVisible);
    	plot.setRangeCrosshairVisible(crosshairVisible);
    	setVerticalAxisTrace(crosshairVisible);
    	setHorizontalAxisTrace(crosshairVisible);
	}

	/* (non-Javadoc)
	 * @see org.jfree.chart.ChartPanel#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
		super.mouseExited(e);
    	XYPlot plot = getChart().getXYPlot();
    	plot.setDomainCrosshairVisible(false);
    	plot.setRangeCrosshairVisible(false);
    	setVerticalAxisTrace(false);
    	setHorizontalAxisTrace(false);
    	SpanMainFrame.getMainFrame().setCursorFreqValue("");
    	SpanMainFrame.getMainFrame().setCursorAmplValue("");
	}
	
	/* (non-Javadoc)
	 * @see org.jfree.chart.ChartPanel#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent e)
	{
		super.mouseMoved(e);

		XYPlot plot = getChart().getXYPlot();
		SpanMainFrame smf = SpanMainFrame.getMainFrame();
		double x = getDomainDataValue(e.getX());
		double y = getRangeDataValue(e.getY());
		boolean showDeltas = false;
		
		// we distinguish two cases:
		// 1. Cursors visible *and* crosshair cursors not at 0/0: display delta
		// 2. Cursors invisible *or* crosshair cursors at 0/0: display absolute value
		if( crosshairVisible 
				&& plot.getDomainCrosshairValue() != 0
				&& plot.getRangeCrosshairValue() != 0	)
		{
			showDeltas = true;
			x -= plot.getDomainCrosshairValue();
			y -= plot.getRangeCrosshairValue();
		}

		String freqLabel = "Frequency";
		String amplLabel = "Amplitude";
		
		// Change frequency to sweep time in zero-span mode
		if( zeroSpanMode )
		{
			smf.setCursorFreqValue(NumericParameter.dur2String(x * xAxisFactor));
			freqLabel = "Sweep Time";
		}
		else
		{
			smf.setCursorFreqValue(NumericParameter.freq2String(x * xAxisFactor));
		}

		String amplString = NumericParameter.dbm2String(y);
		
		if( showDeltas )
		{
			// change units from dBm to dB for delta
			amplString = amplString.replaceAll("dBm", "dB");
			// prepend labels with delta sign
			freqLabel = "\u2206 " + freqLabel;
			amplLabel = "\u2206 " + amplLabel;
		}
		
		smf.setCursorAmplValue(amplString);
		smf.setCursorFreqLabel(freqLabel);
		smf.setCursorAmplLabel(amplLabel);
	}

    /**
	 * Returns the domain data value given a screen coordinate.
	 * @param x the x-coordinate of the screen
	 * @return the corresponding domain data value
	 */
	private double getDomainDataValue(final int x) 
	{
		// get the current domain value
		XYPlot plot = getChart().getXYPlot();
		double xx = plot.getDomainAxis().java2DToValue(
				x, getScreenDataArea(), plot.getDomainAxisEdge()
		);
		return xx;
	}
	
	/**
	 * Returns the range data value given a screen coordinate.
	 * @param y the y-coordinate of the screen
	 * @return the corresponding range data value
	 */
	private double getRangeDataValue(final int y) 
	{
		// get the current range value
		XYPlot plot = getChart().getXYPlot();
		double yy = plot.getRangeAxis().java2DToValue(
				y, getScreenDataArea(), plot.getRangeAxisEdge()
		);
		return yy;
	}
	
	/**
	 * Returns the y value in data space belonging to the given dataset
	 * index and real x value.
	 * @param index
	 * @param xval
	 * @return
	 */
    @SuppressWarnings("unused")
	private double getRealDataValue(int index, double xval)  // currently unused
	{
		XYPlot plot = getChart().getXYPlot();
    	XYSeriesCollection xysc = (XYSeriesCollection)plot.getDataset(index);
		if( xysc != null )
		{
			TraceDataSeries xys = (TraceDataSeries)xysc.getSeries(0);
			if( xys != null )
			{
				return xys.getInterpolatedValue(xval);
			}
		}
		return 0.0;
	}

    /**
     * Returns the y value in screen coordinates belonging to the given
     * y value in data space.
     * @param y
     * @return
     */
	@SuppressWarnings("unused")
	private int getRange2DValue( double y )  // currently unused
	{
		XYPlot plot = getChart().getXYPlot();
		double yy = plot.getRangeAxis().valueToJava2D(
				y,	getScreenDataArea(), plot.getRangeAxisEdge());
		return (int)yy;
	}
   
    /**
     * Draws a horizontal line used to trace the mouse position to the vertical
     * axis.
     *
     * @param y  the y-coordinate of the trace line.
     */
    @SuppressWarnings("unused")
	private void drawVerticalAxisTrace(int y)   // currently unused
    {
        Graphics2D g2 = (Graphics2D) getGraphics();
        Rectangle2D dataArea = getScreenDataArea();

        g2.setXORMode(java.awt.Color.red);
        if (((int) dataArea.getMinY() < y) && (y < (int) dataArea.getMaxY())) {

            if (this.horizontalTraceLine != null) {
                g2.draw(this.horizontalTraceLine);
                this.horizontalTraceLine.setLine(
                    (int) dataArea.getMinX(), y, (int) dataArea.getMaxX(), y
                );
            }
            else {
                this.horizontalTraceLine = new Line2D.Float(
                    (int) dataArea.getMinX(), y, (int) dataArea.getMaxX(), y
                );
            }
            g2.draw(this.horizontalTraceLine);
        }
    }

	/* (non-Javadoc)
	 * @see org.jfree.chart.ChartPanel#zoom(java.awt.geom.Rectangle2D)
	 */
	public void zoom(Rectangle2D selection)
	{
		super.zoom(selection);
		zoomStack.push(selection);
		zoomBackMenuItem.setEnabled(true);
		zoomOutMenuItem.setEnabled(true);
	}
	
	private void zoomBack()
	{
		if( !zoomStack.empty() )
		{		
			zoomStack.pop();
			if( zoomStack.empty() )
			{
				resetAxesRange();
			}
			else
			{
				resetAxesRange();
				super.zoom((Rectangle2D)zoomStack.peek());
			}
		}
		zoomBackMenuItem.setEnabled(isZoomed());
		zoomOutMenuItem.setEnabled(isZoomed());
	}

	private void zoomOut()
	{
		zoomStack.clear();
		resetAxesRange();
		zoomBackMenuItem.setEnabled(false);
		zoomOutMenuItem.setEnabled(false);
	}

	private boolean isZoomed()
	{
		return !zoomStack.empty();
	}

	protected JPopupMenu createPopupMenu(boolean properties, boolean save, boolean print, boolean zoom)
	{
		JPopupMenu menu = super.createPopupMenu(properties, save, print, zoom);
		quickPrintMenuItem = new JMenuItem("Quick print");
		quickPrintMenuItem.setActionCommand(CMD_QUICK_PRINT);
		quickPrintMenuItem.setEnabled(true);
		quickPrintMenuItem.addActionListener(this);
		menu.add(quickPrintMenuItem);
		snapshotMenuItem = new JMenuItem("Snapshot...");
		snapshotMenuItem.setActionCommand(CMD_SNAPSHOT);
		snapshotMenuItem.setEnabled(true);
		snapshotMenuItem.addActionListener(this);
		menu.addSeparator();
		menu.add(snapshotMenuItem);
		zoomBackMenuItem  = new JMenuItem("Zoom back");
		zoomBackMenuItem.setActionCommand(CMD_ZOOM_BACK);
		zoomBackMenuItem.setEnabled(false);
		zoomBackMenuItem.addActionListener(this);
		menu.addSeparator();
		menu.add(zoomBackMenuItem);
		zoomOutMenuItem  = new JMenuItem("Zoom out (100%)");
		zoomOutMenuItem.setActionCommand(CMD_ZOOM_OUT);
		zoomOutMenuItem.setEnabled(false);
		zoomOutMenuItem.addActionListener(this);
		menu.add(zoomOutMenuItem);
		antiAliasCheckBox = new JCheckBoxMenuItem("Anti-aliasing");
		antiAliasCheckBox.setActionCommand(CMD_ANTI_ALIAS);
		antiAliasCheckBox.setSelected(false);
		antiAliasCheckBox.addActionListener(this);
		menu.addSeparator();
		menu.add(antiAliasCheckBox);
		menu.addSeparator();
		// attach to the main frame's action to properly update the state
		toggleChartCheckBox = new JCheckBoxMenuItem(SpanMainFrame.getMainFrame().getToggleChartExpandAction());
		menu.add(toggleChartCheckBox);
		menu.addSeparator();
		menu.add(new JCheckBoxMenuItem(getToggleCursorsAction()));
		return menu;
	}

	/**
	 * Sets/unsets the plot to zero span mode by requesting the {@link SpanMainFrame}
	 * to change the x axis accordingly (frequency / sweep time).
	 * @param zeroSpanMode (true/false)
	 */
	public void setZeroSpanMode(boolean zeroSpanMode)
	{
		if( this.zeroSpanMode != zeroSpanMode )
		{
			this.zeroSpanMode = zeroSpanMode;
			SpanMainFrame.getMainFrame().setZeroSpanMode(zeroSpanMode);
		}
	}

	/* (non-Javadoc)
	 * @see esa.span.client.IMonitoringClient#updateWhenInactive()
	 */
	public boolean updateWhenInactive() {
		// no need to update when iconized
		return false;
	}
	
	@SuppressWarnings("serial")
	public AbstractAction getToggleCursorsAction() {
		if(toggleCursorsAction == null) {
			toggleCursorsAction = new AbstractAction("Show Cursors") {
				private boolean cursorsDisplayed = false;	// remove with Java 6
				public void actionPerformed(ActionEvent e) {
//					setCursorVisible((Boolean) getValue(Action.SELECTED_KEY));	enable with Java 6
					cursorsDisplayed = !cursorsDisplayed;	// remove with Java 6
					setCursorVisible(cursorsDisplayed);		// remove with Java 6
				}
			};
//			toggleCursorsAction.putValue(Action.SELECTED_KEY, false);	enable with Java 6
			toggleCursorsAction.putValue(Action.MNEMONIC_KEY, KeyEvent.VK_C);
		}
		return toggleCursorsAction;
	}

}
